use std::{
    fs,
    path::{Path, PathBuf},
};

use sqlx::{prelude::*, sqlite::SqliteRow};
// use sqlx::{FromRow, sqlite::SqliteRow, sqlx::Row};
use chrono::{DateTime, NaiveDateTime, Utc};

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Sheet {
    pub id: i64,
    pub name: String,
    pub composer_id: i64,
    pub path: PathBuf,
    pub file_size: u64,
    pub file_hash: String,
    pub last_opened: DateTime<Utc>,
}

impl FromRow<'_, SqliteRow> for Sheet {
    fn from_row(row: &SqliteRow) -> sqlx::Result<Self> {
        Ok(Self {
            id: row.try_get("id")?,
            name: row.try_get("name")?,
            composer_id: row.try_get("composer_id")?,
            path: row.try_get::<&str, _>("path")?.into(),
            file_size: row.try_get::<i64, _>("file_size")? as u64,
            file_hash: row.try_get("file_hash")?,
            last_opened: NaiveDateTime::from_timestamp_opt(
                row.try_get::<i64, _>("last_opened")?,
                0,
            )
            .unwrap()
            .and_utc(),
        })
    }
}

#[derive(sqlx::FromRow)]
pub struct Composer {
    pub id: i64,
    pub name: String,
}

impl Sheet {
    pub fn validate_path(&self, path: impl AsRef<Path>) -> std::io::Result<bool> {
        // First compare file size since it is faster than hashing
        let file_size = fs::metadata(path.as_ref())?.len();
        if file_size == self.file_size {
            let file_content = fs::read(path.as_ref())?;
            let file_hash = blake3::hash(&file_content);
            if file_hash.to_string() == self.file_hash {
                return Ok(true);
            }
        }

        Ok(false)
    }
}

impl OrphanFile {
    // TODO: fix duplication
    pub fn validate_path(&self, path: impl AsRef<Path>) -> std::io::Result<bool> {
        // First compare file size since it is faster than hashing
        let file_size = fs::metadata(path.as_ref())?.len();
        if file_size == self.file_size {
            let file_content = fs::read(path.as_ref())?;
            let file_hash = blake3::hash(&file_content);
            if file_hash.to_string() == self.file_hash {
                return Ok(true);
            }
        }

        Ok(false)
    }
}

impl TryFrom<PathBuf> for OrphanFile {
    type Error = std::io::Error;

    fn try_from(path: PathBuf) -> Result<Self, Self::Error> {
        let file_size = fs::metadata(path.as_path())?.len();
        let file_content = fs::read(path.as_path())?;
        let file_hash = blake3::hash(&file_content).to_string();

        Ok(OrphanFile {
            id: -1,
            path,
            file_size,
            file_hash,
            last_opened: DateTime::default(),
        })
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct OrphanFile {
    pub id: i64,
    pub path: PathBuf,
    pub file_size: u64,
    pub file_hash: String,
    pub last_opened: DateTime<Utc>,
}

impl FromRow<'_, SqliteRow> for OrphanFile {
    fn from_row(row: &SqliteRow) -> sqlx::Result<Self> {
        Ok(Self {
            id: row.try_get("id")?,
            path: row.try_get::<&str, _>("path")?.into(),
            file_size: row.try_get::<i64, _>("file_size")? as u64,
            file_hash: row.try_get("file_hash")?,
            last_opened: NaiveDateTime::from_timestamp_opt(
                row.try_get::<i64, _>("last_opened")?,
                0,
            )
            .unwrap()
            .and_utc(),
        })
    }
}
